------------------------------------------------------
-- FactionFriend.lua
------------------------------------------------------
local ADDON_NAME = "GFW_FactionFriend";

FACTION_BAR_COLORS = {
	[1] = {r = 1, g = 0.2, b = 0.15},	-- Hated
	[2] = {r = 0.8, g = 0.3, b = 0.22},	-- Hostile		}
	[3] = {r = 0.75, g = 0.27, b = 0},	-- Unfriendly	} same as default
	[4] = {r = 0.9, g = 0.7, b = 0},	-- Neutral		}
	[5] = {r = 0, g = 0.6, b = 0.1},	-- Friendly		}
	[6] = {r = 0, g = 0.6, b = 0.33},	-- Honored
	[7] = {r = 0, g = 0.7, b = 0.5},	-- Revered
	[8] = {r = 0, g = 0.7, b = 0.7},	-- Exalted
};

FFF_LastBarSwitchTime = 0;
FFF_RecentFactions = {};
FFF_MAX_RECENTS = 10;
FFF_MAX_FACTIONS = 100;	-- infinite loop protection

function FFF_HookTooltip(frame)
	if (frame:GetScript("OnTooltipSetItem")) then
		frame:HookScript("OnTooltipSetItem", FFF_OnTooltipSetItem);
	else
		frame:SetScript("OnTooltipSetItem", FFF_OnTooltipSetItem);
	end
end

function FFF_OnTooltipSetItem(self)
	
	local name, link = self:GetItem();
	if (not link) then return; end
	if (not FFF_Config.NoTooltip and link) then
		local _, _, itemID  = string.find(link, "item:(%d+)");
		itemID = tonumber(itemID);
		
		if (FFF_ReverseCache == nil) then
			FFF_SetupReverseCache();
		end
		
		local itemInfo = FFF_ReverseCache[itemID];
		if (itemInfo) then
			if (type(itemInfo) == "table") then
				FFF_AddTooltipLine(self, FFF_FACTION_TURNIN..":", itemInfo[1]);
				for i = 2, #itemInfo do
					FFF_AddTooltipLine(self, " ", itemInfo[i]);
				end
			else
				FFF_AddTooltipLine(self, FFF_FACTION_TURNIN, itemInfo, 1);
			end
			return true;
		end

	end
	
end

function FFF_AddTooltipLine(tooltip, leftText, faction, isOnlyLine)
	local factionText = faction;
	if (faction == FFF_RACIAL_FACTIONS) then
		faction = UnitFactionGroup("player");
		factionText = faction..FFF_RACIAL_FACTIONS;
	end
	if (FFF_Notes[itemID] and FFF_Notes[itemID][faction]) then
		factionText = factionText.." ("..FFF_Notes[itemID][faction]..")";
	end
	local _, _, standingID = FFF_GetFactionInfoByName(faction);
	if (standingID == nil) then
		standingID = 4;
	end
	local baseColor = FACTION_BAR_COLORS[standingID];
	local color = {};
	for key, value in pairs(baseColor) do 
		color[key] = math.min(1, value * 1.25);
	end
	if (isOnlyLine) then
		tooltip:AddLine(leftText.." "..factionText, color.r, color.g, color.b);
	else
		local c = HIGHLIGHT_FONT_COLOR;
		tooltip:AddDoubleLine(leftText, factionText, c.r, c.g, c.b, color.r, color.g, color.b);
	end
end

function FFF_OnLoad(self)

	hooksecurefunc("ReputationWatchBar_Update", FFF_ReputationWatchBar_Update);
	ReputationWatchBar:HookScript("OnEnter", FFF_ReputationWatchBar_OnEnter);
	ReputationWatchBar:HookScript("OnLeave", FFF_ReputationWatchBar_OnLeave);
	hooksecurefunc("SetWatchedFactionIndex", FFF_SetWatchedFactionIndex);
	
	FFF_HookTooltip(GameTooltip);
	FFF_HookTooltip(ItemRefTooltip);

	FFF_RecentFactions = {};
		
	self:RegisterEvent("UPDATE_FACTION");
	self:RegisterEvent("PLAYER_ENTERING_WORLD");
	self:RegisterEvent("ADDON_LOADED");
	self:RegisterEvent("PLAYER_LEAVING_WORLD");
	self:RegisterEvent("PLAYER_CONTROL_GAINED");
	self:RegisterEvent("CHAT_MSG_COMBAT_FACTION_CHANGE");
	self:RegisterEvent("CHAT_MSG_SYSTEM");
	self:RegisterEvent("ZONE_CHANGED");
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA");
	
	-- Register Slash Commands
	SLASH_FFF1 = "/factionfriend";
	SLASH_FFF2 = "/ff";
	SlashCmdList["FFF"] = function(msg)
		FFF_ShowOptions();
	end
	
end

function FFF_OnEvent(self, event, arg1)
	if ( event == "PLAYER_ENTERING_WORLD" or (event == "ADDON_LOADED" and arg1 == ADDON_NAME)) then
		self:RegisterEvent("BAG_UPDATE");
		self:RegisterEvent("UNIT_INVENTORY_CHANGED");
	elseif( event == "PLAYER_LEAVING_WORLD" ) then
		self:UnregisterEvent("UNIT_INVENTORY_CHANGED");
		self:UnregisterEvent("BAG_UPDATE");
	elseif( event == "PLAYER_CONTROL_GAINED" ) then
		if (FFF_OnTaxi) then
			FFF_OnTaxi = nil;
			FFF_CheckZone();
		end
	elseif ( event == "ZONE_CHANGED_NEW_AREA" or event == "ZONE_CHANGED" ) then
		if (UnitOnTaxi("player")) then
			FFF_OnTaxi = 1;
			return;
		end
		FFF_CheckZone();
	elseif ( event == "CHAT_MSG_COMBAT_FACTION_CHANGE" ) then
		
		if (FFF_FACTION_STANDING_DECREASED == nil) then
			FFF_FACTION_STANDING_DECREASED = GFWUtils.FormatToPattern(FACTION_STANDING_DECREASED);
		end
		local _, _, factionName, amount = string.find(arg1, FFF_FACTION_STANDING_DECREASED);
		if (factionName) then
			FFF_AddToRecentFactions(factionName);
			amount = 0 - amount;
		else
			if (FFF_FACTION_STANDING_INCREASED == nil) then
				FFF_FACTION_STANDING_INCREASED = GFWUtils.FormatToPattern(FACTION_STANDING_INCREASED);
			end
			_, _, factionName, amount = string.find(arg1, FFF_FACTION_STANDING_INCREASED);
			amount = tonumber(amount);
		end
		
		local currentZone = GetRealZoneText();
		local zoneFaction = FFF_ZoneFactions[UnitFactionGroup("player")][currentZone] or FFF_ZoneFactions.Neutral[currentZone];
		if (FFF_Config.NoRepGained) then
			FFF_AddToRecentFactions(factionName);
		elseif (FFF_Config.NoZones) then
			local watchedFaction = GetWatchedFactionInfo();
			if (watchedFaction ~= factionName and amount > 0 and GetTime() - FFF_LastBarSwitchTime > 5) then
				FFF_SetWatchedFaction(factionName);
			end
		else
			if (not zoneFaction or zoneFaction == factionName) then
				local watchedFaction = GetWatchedFactionInfo();
				if (watchedFaction ~= factionName and amount > 0 and GetTime() - FFF_LastBarSwitchTime > 5) then
					FFF_SetWatchedFaction(factionName);
				end
			end
		end
		
		local _, _, standing, min, max, value = FFF_GetFactionInfoByName(factionName);
		if (not standing) then
			-- we don't know about this faction yet
			-- (for a new faction, the gain/loss message often comes before the initial-state message)
			-- queue up the info so we can reprint the message after we get the initial state
			FFF_QueuedMessageFaction = factionName;
			FFF_QueuedMessageOriginal = arg1;
			FFF_QueuedMessageAmount = amount;
			return;
		end
		if (FFF_Config.CountRepeatGains) then
			FFF_PrintReputationMessage(factionName, arg1, amount);
		end
		
	elseif ( event == "CHAT_MSG_SYSTEM" ) then

		if (FFF_FACTION_STANDING_CHANGED == nil) then
			FFF_FACTION_STANDING_CHANGED = GFWUtils.FormatToPattern(FACTION_STANDING_CHANGED);
		end
		_, _, standingText, factionName = string.find(arg1, FFF_FACTION_STANDING_CHANGED);
		if (factionName and factionName == FFF_QueuedMessageFaction) then
			FFF_PrintReputationMessage(FFF_QueuedMessageFaction, FFF_QueuedMessageOriginal, FFF_QueuedMessageAmount);

			if (FFF_QueuedMessageAmount > 0) then
				if (FFF_Config.NoZones) then
					local watchedFaction = GetWatchedFactionInfo();
					if (watchedFaction ~= factionName and GetTime() - FFF_LastBarSwitchTime > 5) then
						FFF_SetWatchedFaction(factionName);
					end
				else
					if (not zoneFaction or zoneFaction == factionName) then
						local watchedFaction = GetWatchedFactionInfo();
						if (watchedFaction ~= factionName and GetTime() - FFF_LastBarSwitchTime > 5) then
							FFF_SetWatchedFaction(factionName);
						end
					end
				end
			end
			FFF_QueuedMessageFaction = nil;
			FFF_QueuedMessageOriginal = nil;
			FFF_QueuedMessageAmount = nil;
		end
		
	elseif ( event == "BAG_UPDATE") then
		FFF_ReputationWatchBar_Update();
	elseif ( event == "UNIT_INVENTORY_CHANGED") then
		if (arg1 == "player") then
			FFF_ReputationWatchBar_Update();
		end
	end
end

function FFF_PrintReputationMessage(factionName, originalMessage, amount)
	local _, _, standing, min, max, value = FFF_GetFactionInfoByName(factionName);
	
	local repeatMessage = "";
	if (tonumber(amount) > 0) then
		-- only show the gains-to-next for gains, not losses
		local nextStanding = standing + 1;
		local repToNext = max - value;
		local gainsToNext = repToNext / amount;

		local nextStandingName, nextColor;
		if (nextStanding < 9) then
			nextStandingName = getglobal("FACTION_STANDING_LABEL"..nextStanding);
			nextColor = GFWUtils.ColorToCode(FACTION_BAR_COLORS[nextStanding]);
		else
			nextStandingName = FFF_MAXIMUM;
			nextColor = GFWUtils.ColorToCode(FACTION_BAR_COLORS[8]);
		end
		nextStandingName = nextColor..nextStandingName..FONT_COLOR_CODE_CLOSE;
		repeatMessage = " "..format(FFF_REPEAT_TURNINS, gainsToNext, nextStandingName);
	end
	
	local color = FACTION_BAR_COLORS[standing];
	DEFAULT_CHAT_FRAME:AddMessage(originalMessage..repeatMessage, color.r, color.g, color.b);
end

local racialFactions = {
	["Alliance"] = { FACTION_DWARF, FACTION_NELF, FACTION_GNOME, FACTION_HUMAN, FACTION_DRAENEI },
	["Horde"] = { FACTION_ORC, FACTION_TAUREN, FACTION_TROLL, FACTION_FORSAKEN, FACTION_BELF },
};
function FFF_SetupReverseCache()
	FFF_ReverseCache = {};
	FFF_Notes = {};
	for faction, quests in pairs(FFF_ItemInfo) do
		local myExcludedFactions = FFF_ExcludedFactions[UnitFactionGroup("player")];
		if (not myExcludedFactions[faction]) then
			for quest, questInfo in pairs(quests) do
				if (not questInfo.otherFactionRequired) then
					for itemID in pairs(questInfo.items) do
						if (FFF_ReverseCache[itemID]) then
							if (type(FFF_ReverseCache[itemID]) == "table") then
								if (GFWTable.KeyOf(FFF_ReverseCache[itemID], faction) == nil) then
									table.insert(FFF_ReverseCache[itemID], faction);
								end
							elseif (FFF_ReverseCache[itemID] ~= faction) then
								local oldValue = FFF_ReverseCache[itemID];
								FFF_ReverseCache[itemID] = {};
								table.insert(FFF_ReverseCache[itemID], oldValue);
								table.insert(FFF_ReverseCache[itemID], faction);
							end
						else
							FFF_ReverseCache[itemID] = faction;
						end
						if (questInfo.note) then
							if (FFF_Notes[itemID] == nil) then
								FFF_Notes[itemID] = {};
							end
							FFF_Notes[itemID][faction] = questInfo.note;
						end
					end
				end
			end
		end
	end
	local myRacialFactions = racialFactions[UnitFactionGroup("player")];
	for itemID, factionInfo in pairs(FFF_ReverseCache) do
		if (type(factionInfo) == "table") then
			local racialSubtracted = GFWTable.Subtract(factionInfo, myRacialFactions);
			if (#factionInfo - #racialSubtracted == #myRacialFactions) then
				if (#racialSubtracted == 0) then
					FFF_ReverseCache[itemID] = FFF_RACIAL_FACTIONS;
				else
					FFF_ReverseCache[itemID] = racialSubtracted;
					table.insert(FFF_ReverseCache[itemID], FFF_RACIAL_FACTIONS);
				end
			end
		end
	end
end

function FFF_GetFactionInfoByName(factionName)	
	local name;
	local factionIndex = 1;
	repeat
		factionInfo = {GetFactionInfo(factionIndex)};
		name = factionInfo[1];
		if (name == factionName) then
			return unpack(factionInfo);
		end
		factionIndex = factionIndex + 1;
	until (not name or factionIndex > FFF_MAX_FACTIONS);
end

function FFF_GetFactionIndex(factionName)
	if (not factionName) then return; end
	local name;
	local factionIndex = 1;
	repeat
		name = GetFactionInfo(factionIndex);
		if (name == factionName) then
			return factionIndex;
		end
		factionIndex = factionIndex + 1;
	until (not name or factionIndex > FFF_MAX_FACTIONS);
end

function FFF_CheckZone()
	if (FFF_Config.NoZones) then return; end
	
	local currentZone = GetRealZoneText();
	local zoneFaction = FFF_ZoneFactions[UnitFactionGroup("player")][currentZone] or FFF_ZoneFactions.Neutral[currentZone];
	if (zoneFaction) then
		local watchedFaction = GetWatchedFactionInfo();
		if (watchedFaction ~= zoneFaction and GetTime() - FFF_LastBarSwitchTime > 5) then
			FFF_SetWatchedFaction(zoneFaction);
		end
	end
end

function FFF_SetWatchedFactionIndex(index)
	-- hook so we can update our recent factions list if you set the watched faction thru another UI
	local name = GetFactionInfo(index);
	if (name) then
		FFF_AddToRecentFactions(name);
	end
end

function FFF_SetWatchedFaction(factionName)
	local index = FFF_GetFactionIndex(factionName);
	if (index and not IsFactionInactive(index)) then
		SetWatchedFactionIndex(index);
		FFF_LastBarSwitchTime = GetTime();
	end
end

function FFF_AddToRecentFactions(factionName)
	local newRecentFactions = {factionName};
	for index = 1, #FFF_RecentFactions do
		if (#newRecentFactions >= FFF_MAX_RECENTS) then
			break;
		end
		if (FFF_RecentFactions[index] ~= factionName) then
			table.insert(newRecentFactions, FFF_RecentFactions[index]);
		end
	end
	FFF_RecentFactions = newRecentFactions;
end

function FFF_IsFactionItem(itemID)
	for factionName, factionInfo in pairs(FFF_ItemInfo) do
		for quest, qInfo in pairs(factionInfo) do
			for id, qtyPerTurnin in pairs(qInfo.items) do
				if (id == itemID) then 
					return true; 
				end
			end
		end
	end
end

function FFF_GetItemCount(itemID, includeBank)
	if (FFF_FakeItemCount and FFF_FakeItemCount[itemID]) then
		-- useful for debugging
		return FFF_FakeItemCount[itemID];
	end
	return GetItemCount(itemID, includeBank);
end

FFF_SpecialItems = {
	[12844] = { n=FFF_AD_TOKEN, q=2 },
	[19858] = { n=FFF_ZG_TOKEN, q=2 },
}
function FFF_ItemLink(itemID)
	local _, link = GetItemInfo(itemID);
	if (link) then return link; end
	
	local info = FFF_SpecialItems[itemID];
	if (info) then
		local _, _, _, colorCode = GetItemQualityColor(info.q);
		return colorCode .. info.n .. FONT_COLOR_CODE_CLOSE;
	end
end

FFF_PointsPerStanding = {
	[1] = 36000,	-- Hated
	[2] = 3000,		-- Hostile
	[3] = 3000,		-- Unfriendly
	[4] = 3000,		-- Neutral
	[5] = 6000,		-- Friendly
	[6] = 12000,	-- Honored
	[7] = 21000,	-- Revered
	[8] = 1000,
};
FFF_AbsMinForStanding = {
	[1] = 0 - FFF_PointsPerStanding[3] - FFF_PointsPerStanding[2] - FFF_PointsPerStanding[1],
	[2] = 0 - FFF_PointsPerStanding[3] - FFF_PointsPerStanding[2],
	[3] = 0 - FFF_PointsPerStanding[3],
	[4] = 0,
	[5] = FFF_PointsPerStanding[4],
	[6] = FFF_PointsPerStanding[4] + FFF_PointsPerStanding[5],
	[7] = FFF_PointsPerStanding[4] + FFF_PointsPerStanding[5] + FFF_PointsPerStanding[6],
	[8] = FFF_PointsPerStanding[4] + FFF_PointsPerStanding[5] + FFF_PointsPerStanding[6] + FFF_PointsPerStanding[7],
	[9] = FFF_PointsPerStanding[4] + FFF_PointsPerStanding[5] + FFF_PointsPerStanding[6] + FFF_PointsPerStanding[7] + FFF_PointsPerStanding[8],
};

function FFF_StandingForValue(value)
	value = math.min(value, FFF_AbsMinForStanding[9] - 1);	-- cap at max Exalted
	value = math.max(value, FFF_AbsMinForStanding[1]);	-- and min Hated
	for standingID = 1, 8 do
		if (value >= FFF_AbsMinForStanding[standingID] and value < FFF_AbsMinForStanding[standingID+1]) then
			local pointsInto = value - FFF_AbsMinForStanding[standingID];
			return standingID, pointsInto, FFF_PointsPerStanding[standingID];
		end
	end
end

function FFF_GetWatchedFactionPotential(withReport)
	return FFF_GetFactionPotential(nil, withReport)
end

function FFF_GetFactionPotential(faction, withReport)
	local factionName, standing, min, max, value;
	if (faction == nil) then
		factionName, standing, min, max, value = GetWatchedFactionInfo();
	elseif (type(faction) == "string") then
		factionName = faction;
		_, _, standing, min, max, value = FFF_GetFactionInfoByName(factionName);
		if (standing == nil) then return 0; end
	elseif (type(faction) == "number") then
		_, _, standing, min, max, value = GetFactionInfo(faction);
		if (standing == nil) then return 0; end
	else
		error("bad argument #1 to 'FFF_GetFactionPotential' (expected number, string, or nil, got "..type(faction)..")", 2);
	end	
		
	local factionInfo = FFF_ItemInfo[factionName];
	if (factionInfo == nil) then return 0; end
	
	local totalPotential = 0;
	local reportLines;
	if (withReport) then
		reportLines = {};
	end
	local multiplier = 1;
	local _, race = UnitRace("player");
	if (race == "Human") then
		multiplier = 1.1;
	end
	local itemsAccountedFor = {};
	local itemsCreated = {};
	for quest, qInfo in GFWTable.PairsByKeys(factionInfo, function(a,b) return a > b end) do
		local meetsRequirements;
		
		meetsRequirements = (standing >= (qInfo.minStanding or 1));
		meetsRequirements = meetsRequirements and (standing <= (qInfo.maxStanding or 8));

		if (meetsRequirements and qInfo.otherFactionRequired) then
			local _, _, standingID = FFF_GetFactionInfoByName(qInfo.otherFactionRequired.faction);
			if (standingID) then
				meetsRequirements = meetsRequirements and (standingID >= (qInfo.otherFactionRequired.minStanding or 1));
				meetsRequirements = meetsRequirements and (standingID <= (qInfo.otherFactionRequired.maxStanding or 8));
			end
		end
		
		if (meetsRequirements) then
			local potentialValue = 0;
			
			-- first, figure out how many turnins' worth we have of each item the quest requres
			local turninCounts = {};
			local reportItemLines = {};
			for itemID, qtyPerTurnin in pairs(qInfo.items) do
				local countIncludingBank = FFF_GetItemCount(itemID, true);
				local created = itemsCreated[itemID] or 0;
				local alreadyCounted = itemsAccountedFor[itemID] or 0;
				local turnins = math.floor((countIncludingBank + created - alreadyCounted) / qtyPerTurnin);
				table.insert(turninCounts, turnins);
			end
			
			-- if a quest requires multiple item, the number of turnins we can actually make
			-- is decided by whichever item we have the fewest turnins' worth of
			local numTurnins;
			if (table.getn(turninCounts) == 1) then
				numTurnins = turninCounts[1];
			elseif (table.getn(turninCounts) > 1) then
				numTurnins = math.min(unpack(turninCounts));
			else
				numTurnins = 0;
			end
			
			-- some turnins create items which can be used for later "turnins"
			if (qInfo.creates) then
				for createdID, qtyCreated in pairs(qInfo.creates) do
					itemsCreated[createdID] = (itemsCreated[createdID] or 0) + (qtyCreated * numTurnins);
				end
			end

			-- now that we know how many we're actually using of each item for this quest,
			-- account for it so they don't get used again for another quest,
			-- and include it in the report/tooltip
			for itemID, qtyPerTurnin in pairs(qInfo.items) do
				itemsAccountedFor[itemID] = numTurnins * qtyPerTurnin;
				
				if (withReport and numTurnins > 0) then
					local itemsForTurnin = itemsAccountedFor[itemID];
					local itemLink = FFF_ItemLink(itemID) or string.format(FFF_UNKNOWN_ITEM, itemID);
					local lineItem = string.format(FFF_REPORT_LINE_ITEM, itemsForTurnin, itemLink);
					local inBags = FFF_GetItemCount(itemID, false);
					local inBank = FFF_GetItemCount(itemID, true) - inBags;

					local lineItemAdditions = {};
					if (inBank > 0 and inBags < itemsForTurnin) then
						table.insert(lineItemAdditions, string.format(FFF_COUNT_IN_BANK, math.min(inBank, itemsForTurnin)));
					end
					local created = itemsCreated[itemID];
					if (created and inBank == 0 and created == itemsForTurnin) then
						table.insert(lineItemAdditions, FFF_ALL_CREATED);
					elseif (created and created > 0 and inBags + inBank < itemsForTurnin) then
						table.insert(lineItemAdditions, string.format(FFF_COUNT_CREATED, math.min(created, itemsForTurnin)));
					end
					if (#lineItemAdditions > 0) then
						lineItem = lineItem ..GRAY_FONT_COLOR_CODE.. " (".. table.concat(lineItemAdditions, ", ") .. ")" .. FONT_COLOR_CODE_CLOSE;
					end
					table.insert(reportItemLines, lineItem);
				end
			end
			
			-- Now we can figure how much rep we stand to gain from this quest...
			potentialValue = numTurnins * qInfo.value * multiplier;
			-- ... but we may need to adjust it in case it'd push us up to a new standing
			local potentialStanding = standing;
			local pointsIntoStanding;
			if (potentialValue > max) then
				local potentialTotal = value + potentialValue;
				potentialStanding, pointsIntoStanding = FFF_StandingForValue(potentialTotal);
			end
			if (qInfo.maxStanding and potentialStanding >= qInfo.maxStanding) then
				local maxValue = qInfo.maxValue or FFF_PointsPerStanding[qInfo.maxStanding];
				local absMaxValue = FFF_AbsMinForStanding[qInfo.maxStanding] + maxValue;
				if (value >= absMaxValue) then
					potentialValue = 0;
				end
				local maxPotential = absMaxValue - value;
				potentialValue = math.min(potentialValue, maxPotential);
			end
			
			-- Finally, throw what we've figured out into the report/tooltip
			if (potentialValue > 0 and withReport) then
				local reportLineHeader = string.format(FFF_REPORT_NUM_POINTS, potentialValue);
				if (numTurnins > 1 and not qInfo.useItem) then
					reportLineHeader = reportLineHeader  ..GRAY_FONT_COLOR_CODE..string.format(FFF_REPORT_NUM_TURNINS, numTurnins)..FONT_COLOR_CODE_CLOSE;
				end
				reportLineHeader = reportLineHeader ..":";
				table.insert(reportLines, {[reportLineHeader] = reportItemLines});
			end
			
			totalPotential = totalPotential + potentialValue;
		end
	end
	return totalPotential, reportLines;

end

function FFF_ReputationWatchBar_Update(newLevel)
	
	local name, standing, min, max, value = GetWatchedFactionInfo();
	if (not name) then return; end
	
	if (name ~= FFF_RecentFactions[1]) then
		FFF_AddToRecentFactions(name);
	end
	
	local potential = FFF_GetWatchedFactionPotential();
	local totalValue = value + potential;
	local tickSet = ((totalValue - min) / (max - min)) * ReputationWatchStatusBar:GetWidth();
	local tickSet = math.max(tickSet, 0);
	local tickSet = math.min(tickSet, ReputationWatchStatusBar:GetWidth());
	FFF_ReputationTick:ClearAllPoints();
	if (potential == 0 or FFF_Config.NoShowPotential) then
	    FFF_ReputationTick:Hide();
	    FFF_ReputationExtraFillBarTexture:Hide();
	else
		if (totalValue > FFF_AbsMinForStanding[standing+1]) then 
			FFF_ReputationTick:Hide();
		else
		    FFF_ReputationTick:Show();
		end
	    FFF_ReputationTick:SetPoint("CENTER", "ReputationWatchStatusBar", "LEFT", tickSet, 0);
	    FFF_ReputationExtraFillBarTexture:Show();
	    FFF_ReputationExtraFillBarTexture:SetPoint("TOPRIGHT", "ReputationWatchStatusBar", "TOPLEFT", tickSet, 0);
	    local color = FACTION_BAR_COLORS[standing];
		FFF_ReputationTickHighlight:SetVertexColor(color.r, color.g, color.b);
		if (totalValue > FFF_AbsMinForStanding[standing+1]) then 
			local potentialStanding = FFF_StandingForValue(totalValue);
			color = FACTION_BAR_COLORS[potentialStanding];
		    FFF_ReputationExtraFillBarTexture:SetVertexColor(color.r, color.g, color.b, 0.25);
		else
		    FFF_ReputationExtraFillBarTexture:SetVertexColor(color.r, color.g, color.b, 0.15);
		end
	end

end

function FFF_ReputationTick_Tooltip()
	local x,y;
	x,y = FFF_ReputationTick:GetCenter();
	if ( FFF_ReputationTick:IsVisible() ) then
		FFF_ReputationTick:LockHighlight();
		if ( x >= ( GetScreenWidth() / 2 ) ) then
			GameTooltip:SetOwner(FFF_ReputationTick, "ANCHOR_LEFT");
		else
			GameTooltip:SetOwner(FFF_ReputationTick, "ANCHOR_RIGHT");
		end
	else
		GameTooltip_SetDefaultAnchor(GameTooltip, UIParent);
	end
	
	local potential, reportLines = FFF_GetWatchedFactionPotential(1);
	local tooltipText = string.format(FFF_REPUTATION_TICK_TOOLTIP, potential);
	local name, standing, min, max, value = GetWatchedFactionInfo();
	local color = FACTION_BAR_COLORS[standing];

	-- First line: name of faction and label for current standing
	-- e.g. "The Aldor: Friendly"
	local standingName = getglobal("FACTION_STANDING_LABEL"..standing);
	if (UnitSex("player") == 3) then
		standingName = getglobal("FACTION_STANDING_LABEL"..standing.."_FEMALE");
	end
	GameTooltip:SetText(name..": "..standingName, color.r, color.g, color.b);

	if (potential == 0) then return; end

	-- Report lines summarizing faction available from turnins
	-- e.g. "500 points (2 turnins): 20x[Mark of Sargeras] (13 in bank)"
	GameTooltip:AddLine(tooltipText);
	for _, reportLine in pairs(reportLines) do
		for reportHeader, itemLines in pairs(reportLine) do
			local r, g, b = HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b;
			GameTooltip:AddDoubleLine(reportHeader, itemLines[1], r, g, b, r, g, b);
			if (table.getn(itemLines) > 1) then
				for i = 2, table.getn(itemLines) do
					GameTooltip:AddDoubleLine(" ", itemLines[i], r, g, b, r, g, b);
				end
			end
		end
	end
	
	-- Last line (preceded by blank): standing after turnins (if different from current standing)
	-- e.g. "After turnins: Honored (100/12000)"
	local potentialTotal = potential + value;
	local potentialStanding, pointsInto, localMax = FFF_StandingForValue(potentialTotal);
	if (potentialStanding ~= standing) then
		GameTooltip:AddLine(" ");
		local potentialStandingName = getglobal("FACTION_STANDING_LABEL"..potentialStanding);
		if (UnitSex("player") == 3) then
			potentialStandingName = getglobal("FACTION_STANDING_LABEL"..potentialStanding.."_FEMALE");
		end
		local summary = string.format(FFF_AFTER_TURNINS_INFO, potentialStandingName, pointsInto, localMax);
		local c1 = HIGHLIGHT_FONT_COLOR;
		local c2 = FACTION_BAR_COLORS[potentialStanding];
		GameTooltip:AddDoubleLine(FFF_AFTER_TURNINS_LABEL, summary, c1.r, c1.g, c1.b, c2.r, c2.g, c2.b);
	end
	
	GameTooltip:Show();

end

function FFF_ReputationWatchBar_OnEnter()
	if (FFF_Config.NoShowPotential) then return; end
	FFF_ReputationTick_Tooltip();
	FFF_ShowingTooltip = true;
end

function FFF_ReputationWatchBar_OnLeave()
	FFF_ReputationTick:UnlockHighlight();
	if (FFF_ShowingTooltip) then
		GameTooltip:Hide();
		FFF_ShowingTooltip = false;
	end
end

------------------------------------------------------
-- ReputationWatchBar menu 
------------------------------------------------------

function FFF_ReputationWatchBar_OnClick(button)
	if (button == "RightButton") then
		ToggleDropDownMenu(1, nil, FFF_MenuDropDown, "cursor", 0, 0, "TOP");
		PlaySound("igMainMenuOptionCheckBoxOn");
	end
end

function FFF_BuildFactionTree()

	-- first, expand all headers so that we can see the real hierarchy
	-- but remember collapsed state so we can preserve the player's reputation UI
	FFF_CollapsedHeaders = {};
	for factionIndex = 1, GetNumFactions() do
		local name, _, _, _, _, _, _, _, isHeader, isCollapsed = GetFactionInfo(factionIndex);
		if (isHeader and isCollapsed) then
			table.insert(FFF_CollapsedHeaders, name);
			ExpandFactionHeader(factionIndex);
		end
	end
	
	local factionTree = {};
	local name, _, standingID, _, _, _, _, _, isHeader, _, hasRep, isWatched, isChild;
	local factionIndex, treeIndex = 1, 1;
	local currentMainHeader, currentSubHeader, childIndex;
	repeat
		name, _, standingID, _, _, _, _, _, isHeader, _, hasRep, isWatched, isChild = GetFactionInfo(factionIndex);
		if (isHeader and name == FACTION_INACTIVE) then
			break;
		end
		if (name) then
			if (isHeader and not isChild) then
				-- top level headers
				factionTree[treeIndex] = {name=name, children={}};
				currentMainHeader = factionTree[treeIndex].children;
				treeIndex = treeIndex + 1;
				childIndex = 1;
			elseif (isHeader and isChild) then
				-- 2nd level headers
				currentMainHeader[childIndex] = {name=name, children={}, standingID=standingID, isWatched=isWatched, hasRep=hasRep};
				currentSubHeader = currentMainHeader[childIndex].children;
				childIndex = childIndex + 1;
				grandchildIndex = 1;
			elseif (isChild) then
				-- 3rd level items (children of 2nd level headers)
				currentSubHeader[grandchildIndex] = {name=name, standingID=standingID, isWatched=isWatched};
				grandchildIndex = grandchildIndex + 1;
			else
				-- 2nd level items (children of top level headers)
				currentMainHeader[childIndex] = {name=name, standingID=standingID, isWatched=isWatched};
				childIndex = childIndex + 1;
			end
		end		
		factionIndex = factionIndex + 1;
	until (not name or factionIndex > FFF_MAX_FACTIONS);
	
	-- TODO: restore collapsed headers

	return factionTree;
end

-- for debug only
function FFF_PrintFactionTree()
	local tree = FFF_BuildFactionTree();
	for _, header in pairs(tree) do
		print(header.name);
		for _, item in pairs(header.children) do
			if (item.children) then
				-- it's a subheader
				print(" ", item.name, item.standingID, item.isWatched, item.hasRep);
				for _, subItem in pairs(item.children) do
					print(" ", " ", subItem.name,":", subItem.standingID, subItem.isWatched);
				end
			else
				-- it's a faction
				print(" ", item.name,":", item.standingID, item.isWatched);
			end
		end
	end				
end

function FFF_CountActiveFactions(tree)
	local count = 0;
	for _, header in pairs(tree) do
		if (#header.children > 0) then
			count = count + 1;
			for _, item in pairs(header.children) do
				if (item.children) then
					if (#item.children > 0) then
						count = count + 1;
						for _, subItem in pairs(item.children) do
							count = count + 1;
						end
					end
				else
					count = count + 1;
				end
			end
		end
	end		
	return count;		
end

function FFF_CreateFactionMenuItem(name, standingID, isWatched, gender)
--[[
	local baseColor = FACTION_BAR_COLORS[standingID];
	local color = {};
	for key, value in pairs(baseColor) do 
		color[key] = math.min(1, value * 1.25);
	end
	local standingLabel = GetText("FACTION_STANDING_LABEL"..standingID, gender);
]]
	local info = UIDropDownMenu_CreateInfo();
	--info.text = GFWUtils.ColorText("("..standingLabel..")", color).." "..name;
	info.text = name;
	info.checked = isWatched;
	if (not isWatched) then
		info.func = function()
			FFF_SetWatchedFaction(name);
			CloseDropDownMenus();
		end;
	end
	return info;
end

function FFF_CreateSeparatorMenuItem()
	local info = UIDropDownMenu_CreateInfo();
	info.disabled = 1;
	info.text = "";
	return info;
end

function FFF_MenuDropDown_OnLoad(self)
	UIDropDownMenu_Initialize(self, FFF_MenuDropDown_Initialize, "MENU");
end

function FFF_MenuDropDown_Initialize(dropDown)
	
	local tree = FFF_BuildFactionTree();
	local gender = UnitSex("player");
	local info;
	
	dropDown.displayMode = "MENU";
	if (FFF_CountActiveFactions(tree) > 35) then
		-- we'll use submenus
		
		if ( UIDROPDOWNMENU_MENU_LEVEL == 2 ) then
			for _, header in pairs(tree) do
				if (header.name == UIDROPDOWNMENU_MENU_VALUE) then
					for _, item in pairs(header.children) do
					
						if (item.children) then
							-- it's a subheader
							if (#item.children > 0) then

								info = FFF_CreateFactionMenuItem(item.name, item.standingID, item.isWatched, gender);
								info.icon = "Interface\\AddOns\\GFW_FactionFriend\\StandingIcons\\"..item.standingID;
								if (not item.hasRep) then						
									info.text = item.name;
									info.isTitle = 1;
									info.icon = nil;
								end
								UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);

								for _, subItem in pairs(item.children) do
									info = FFF_CreateFactionMenuItem(subItem.name, subItem.standingID, subItem.isWatched, gender);
									info.text = "    "..info.text;
									info.icon = "Interface\\AddOns\\GFW_FactionFriend\\StandingIcons\\"..subItem.standingID;
									UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
								end

							end

						else
							-- it's a faction
							info = FFF_CreateFactionMenuItem(item.name, item.standingID, item.isWatched, gender);
							info.text = "    "..info.text;
							info.icon = "Interface\\AddOns\\GFW_FactionFriend\\StandingIcons\\"..item.standingID;
							UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
						end
					end
					break;
				end
			end
			return;
		end
	
		for _, header in pairs(tree) do
		
			if (header.children and #header.children > 0) then
				info = UIDropDownMenu_CreateInfo();
				info.text = header.name;
				info.value = header.name;
				info.hasArrow = 1;
				UIDropDownMenu_AddButton(info);
			end
		end
		
		info = FFF_CreateSeparatorMenuItem();
		UIDropDownMenu_AddButton(info);				

		info = UIDropDownMenu_CreateInfo();
		info.text = FFF_RECENT_FACTIONS;
		info.isTitle = true;
		info.notCheckable = 1;
		UIDropDownMenu_AddButton(info);
		
		local name, standingID = GetWatchedFactionInfo();
		info = FFF_CreateFactionMenuItem(name, standingID, 1, gender);
		info.icon = "Interface\\AddOns\\GFW_FactionFriend\\StandingIcons\\"..standingID;
		UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
		
		for index = 2, #FFF_RecentFactions do
			local name = FFF_RecentFactions[index];
			local _, _, standingID = FFF_GetFactionInfoByName(name);
			if (standingID) then
				info = FFF_CreateFactionMenuItem(name, standingID, nil, gender);
				info.icon = "Interface\\AddOns\\GFW_FactionFriend\\StandingIcons\\"..standingID;
				UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
			end
		end

		info = FFF_CreateSeparatorMenuItem();
		UIDropDownMenu_AddButton(info);				
		
	else
		-- few enough factions that we can do it all in one menu
		
		for _, header in pairs(tree) do
		
			info = UIDropDownMenu_CreateInfo();
			info.text = LIGHTYELLOW_FONT_COLOR_CODE.."- "..header.name.." -"..FONT_COLOR_CODE_CLOSE;
			info.justifyH = "CENTER";
			info.isTitle = true;
			info.notCheckable = 1;
			UIDropDownMenu_AddButton(info);
						
			for _, item in pairs(header.children) do
				if (item.children) then
					-- it's a subheader
					if (#item.children > 0) then
					
						info = FFF_CreateFactionMenuItem(item.name, item.standingID, item.isWatched, gender);
						if (item.hasRep) then
							info.icon = "Interface\\AddOns\\GFW_FactionFriend\\StandingIcons\\"..item.standingID;
						else
							info.text = item.name;
							info.isTitle = true;
						end
						UIDropDownMenu_AddButton(info);
					
						for _, subItem in pairs(item.children) do
							info = FFF_CreateFactionMenuItem(subItem.name, subItem.standingID, subItem.isWatched, gender);
							info.icon = "Interface\\AddOns\\GFW_FactionFriend\\StandingIcons\\"..subItem.standingID;
							info.text = "      "..info.text;
							UIDropDownMenu_AddButton(info);
						end
					
					end
					
				else
					-- it's a faction
					info = FFF_CreateFactionMenuItem(item.name, item.standingID, item.isWatched, gender);
					info.icon = "Interface\\AddOns\\GFW_FactionFriend\\StandingIcons\\"..item.standingID;
					UIDropDownMenu_AddButton(info);
				end
			end
			
			info = FFF_CreateSeparatorMenuItem();
			UIDropDownMenu_AddButton(info);
			
		end				
		
	end
		
	info = UIDropDownMenu_CreateInfo();
	info.text = FFF_SHOW_REPUTATION_PANE;
	info.func = function() 
		ToggleCharacter("ReputationFrame"); 
	end;
	info.notCheckable = 1;
	UIDropDownMenu_AddButton(info);

	info = UIDropDownMenu_CreateInfo();
	info.text = FFF_SHOW_OPTIONS;
	info.func = function()
		FFF_ShowOptions();
	end;
	info.notCheckable = 1;
	UIDropDownMenu_AddButton(info);
	
end

local oldFunc = ReputationWatchBar:GetScript("OnMouseDown");
if (oldFunc) then
	ReputationWatchBar:SetScript("OnMouseDown", function() oldFunc(); FFF_ReputationWatchBar_OnClick(arg1); end);
else
	ReputationWatchBar:SetScript("OnMouseDown", function() FFF_ReputationWatchBar_OnClick(arg1); end);
end

------------------------------------------------------
-- Dongle & GFWOptions stuff
------------------------------------------------------

GFW_FactionFriend = {};
local GFWOptions = DongleStub("GFWOptions-1.0");

local function buildOptionsUI(panel)

	GFW_FactionFriend.optionsText = {
		Zones = FFF_OPTION_ZONES,
		RepGained = FFF_OPTION_REP_GAINED,
		CountRepeatGains = FFF_OPTION_REPEAT_GAINS,
		ShowPotential = FFF_OPTION_SHOW_POTENTIAL,
		Tooltip = FFF_OPTION_TOOLTIP,
	};
	
	local s, widget, lastWidget;

	widget = panel:CreateCheckButton("ShowPotential", true);
	widget:SetPoint("TOPLEFT", panel.contentAnchor, "BOTTOMLEFT", -2, -16);
	lastWidget = widget;
	
	widget = panel:CreateCheckButton("Tooltip", true);
	widget:SetPoint("TOPLEFT", lastWidget, "BOTTOMLEFT", 0, -2);
	lastWidget = widget;

	widget = panel:CreateCheckButton("CountRepeatGains", false);
	widget:SetPoint("TOPLEFT", lastWidget, "BOTTOMLEFT", 0, -2);
	lastWidget = widget;
	
	s = panel:CreateFontString("FFF_OptionsPanel_SwitchHeader", "ARTWORK", "GameFontNormal");
	s:SetPoint("TOPLEFT", lastWidget, "BOTTOMLEFT", 2, -16);
	s:SetText(FFF_OPTIONS_SWITCHBAR);
	lastWidget = s;

	widget = panel:CreateCheckButton("Zones", true);
	widget:SetPoint("TOPLEFT", lastWidget, "BOTTOMLEFT", -2, -2);
	lastWidget = widget;

	widget = panel:CreateCheckButton("RepGained", true);
	widget:SetPoint("TOPLEFT", lastWidget, "BOTTOMLEFT", 0, -2);
	lastWidget = widget;

	s = panel:CreateFontString("FFF_OptionsPanel_SwitchHeader", "ARTWORK", "GameFontHighlight");
	s:SetPoint("TOPLEFT", lastWidget, "BOTTOMLEFT", 24, -24);
	s:SetPoint("BOTTOMRIGHT", panel, -24, 24);
	s:SetJustifyH("LEFT");
	s:SetJustifyV("TOP");
	s:SetText(FFF_OPTIONS_TIPS);

end

function FFF_ShowOptions()
	local openFunc = InterfaceOptionsFrame_OpenToCategory or InterfaceOptionsFrame_OpenToFrame;
	openFunc(FFF_OptionsPanel);
end

function GFW_FactionFriend:Initialize()
	self.defaults = { 
		profile = {
			NoZones = false,
			NoRepGained = false,
			NoShowPotential = false,
			NoTooltip = false,
			CountRepeatGains = false,
		}
	};
	self.db = self:InitializeDB("GFW_FactionFriendDB", self.defaults);
	FFF_Config = self.db.profile;
end

function GFW_FactionFriend:Enable()
	-- conditionalize 2.4 stuff for now so we can run on 2.3
	if (InterfaceOptions_AddCategory) then
		GFWOptions:CreateMainPanel("GFW_FactionFriend", "FFF_OptionsPanel");
		FFF_OptionsPanel.BuildUI = buildOptionsUI;
	end
end

function GFW_FactionFriend:OptionsChanged()
	FFF_ReputationWatchBar_Update();
	if (FFF_Config.CountRepeatGains) then
		if (GFWTable.KeyOf(DEFAULT_CHAT_FRAME.messageTypeList, "COMBAT_FACTION_CHANGE")) then
			FFF_Config.RepMessagesInDefaultChatFrame = 1;
		end
		ChatFrame_RemoveMessageGroup(DEFAULT_CHAT_FRAME, "COMBAT_FACTION_CHANGE");
	else
		if (FFF_Config.RepMessagesInDefaultChatFrame) then
			ChatFrame_AddMessageGroup(DEFAULT_CHAT_FRAME, "COMBAT_FACTION_CHANGE");
		end
	end
end

GFW_FactionFriend = DongleStub("Dongle-1.2"):New("GFW_FactionFriend", GFW_FactionFriend);

